{
  "type": "record",
  "name": "CorpUserCredentials",
  "namespace": "com.linkedin.pegasus2avro.identity",
  "doc": "Corp user credentials",
  "fields": [
    {
      "name": "salt",
      "type": "string",
      "doc": "Salt used to hash password"
    },
    {
      "name": "hashedPassword",
      "type": "string",
      "doc": "Hashed password generated by concatenating salt and password, then hashing"
    },
    {
      "name": "passwordResetToken",
      "type": [
        "null",
        "string"
      ],
      "doc": "Optional token needed to reset a user's password. Can only be set by the admin.",
      "default": null
    },
    {
      "name": "passwordResetTokenExpirationTimeMillis",
      "type": [
        "null",
        "long"
      ],
      "doc": "When the password reset token expires.",
      "default": null
    }
  ],
  "Aspect": {
    "EntityUrns": [
      "com.linkedin.pegasus2avro.common.CorpuserUrn"
    ],
    "name": "corpUserCredentials"
  }
}