{
  "type": "record",
  "name": "CorpUserSettings",
  "namespace": "com.linkedin.pegasus2avro.identity",
  "doc": "Settings that a user can customize through the datahub ui",
  "fields": [
    {
      "name": "appearance",
      "type": {
        "type": "record",
        "name": "CorpUserAppearanceSettings",
        "doc": "Settings for a user around the appearance of their DataHub UI",
        "fields": [
          {
            "name": "showSimplifiedHomepage",
            "type": [
              "null",
              "boolean"
            ],
            "doc": "Flag whether the user should see a homepage with only datasets, charts and dashboards. Intended for users\nwho have less operational use cases for the datahub tool.",
            "default": null
          }
        ]
      },
      "doc": "Settings for a user around the appearance of their DataHub U"
    }
  ],
  "Aspect": {
    "name": "corpUserSettings"
  }
}