{
  "type": "record",
  "name": "GlossaryTermInfo",
  "namespace": "com.linkedin.pegasus2avro.glossary",
  "doc": "Properties associated with a GlossaryTerm",
  "fields": [
    {
      "name": "customProperties",
      "type": {
        "type": "map",
        "values": "string"
      },
      "doc": "Custom property bag.",
      "default": {},
      "Searchable": {
        "/*": {
          "queryByDefault": true
        }
      }
    },
    {
      "name": "id",
      "type": [
        "null",
        "string"
      ],
      "doc": "Optional id for the term",
      "default": null,
      "Searchable": {
        "fieldType": "TEXT_PARTIAL"
      }
    },
    {
      "name": "name",
      "type": [
        "null",
        "string"
      ],
      "doc": "Display name of the term",
      "default": null,
      "Searchable": {
        "boostScore": 10.0,
        "enableAutocomplete": true,
        "fieldType": "TEXT_PARTIAL"
      }
    },
    {
      "name": "definition",
      "type": "string",
      "doc": "Definition of business term.",
      "Searchable": {}
    },
    {
      "name": "parentNode",
      "type": [
        "null",
        "string"
      ],
      "doc": "Parent node of the glossary term",
      "default": null,
      "Relationship": {
        "entityTypes": [
          "glossaryNode"
        ],
        "name": "IsPartOf"
      },
      "Searchable": {
        "fieldName": "parentNode",
        "fieldType": "URN",
        "hasValuesFieldName": "hasParentNode"
      },
      "java": {
        "class": "com.linkedin.pegasus2avro.common.urn.GlossaryNodeUrn"
      }
    },
    {
      "name": "termSource",
      "type": "string",
      "doc": "Source of the Business Term (INTERNAL or EXTERNAL) with default value as INTERNAL",
      "Searchable": {
        "fieldType": "KEYWORD"
      }
    },
    {
      "name": "sourceRef",
      "type": [
        "null",
        "string"
      ],
      "doc": "External Reference to the business-term",
      "default": null,
      "Searchable": {
        "fieldType": "KEYWORD"
      }
    },
    {
      "name": "sourceUrl",
      "type": [
        "null",
        "string"
      ],
      "doc": "The abstracted URL such as https://spec.edmcouncil.org/fibo/ontology/FBC/FinancialInstruments/FinancialInstruments/CashInstrument.",
      "default": null,
      "java": {
        "class": "com.linkedin.pegasus2avro.common.url.Url",
        "coercerClass": "com.linkedin.pegasus2avro.common.url.UrlCoercer"
      }
    },
    {
      "name": "rawSchema",
      "type": [
        "null",
        "string"
      ],
      "doc": "Schema definition of the glossary term",
      "default": null,
      "deprecated": true
    }
  ],
  "Aspect": {
    "name": "glossaryTermInfo"
  }
}